/*
 * copap.c: 
 * 
 * C-MEX file for solving the circular order preserving assignment problem 
 *          
 * USAGE
 *
 * [P,min_cost]=copap(C,ell,dum_cost,delta)
 *
 * Fourth argument is optional (default = 1)
 * 
 * INPUTS
 *
 * C = m x n cost matrix 
 * ell: require matching of at least this size
 * dum_cost = cost of assigning a point to a dummy point
 * delta = amount to increment shift when circularly shifting C
 * 
 * OUTPUTS
 *
 * P = 1 x N permutation vector: P(i) = row assigned to i'th column
 * min_cost = cost of optimal assignment
 *
 *
 * Created 8/1/2002 by Clay Scott
 * Memory management functions borrowed from Don Waagen
 * Last uptated 6/13/2005 by Clay Scott
 * Author contact: cscott@rice.edu
 *
 */ 

#include "mex.h"
#include <stdio.h>
#include <stdlib.h>

#define HUGE 1000000000000    /* very large number */

/* 
 * Global variables.  They are global so that we don't have to 
 * allocate/deallocate space for them with every call of lopap.
 */

double ***L;
int ***I;
int ***J; 
int **x;
int minM;
int maxM;
double *min_cost;
double *u;
int *v;
double *Ttmp;
int *Ibest;
int *Jbest;
int *filledrows;
int *filledcols;

/*
 * Memory allocation utilities
 */ 

/* ********************************************************************
// function: mxCalloc3d
// 
// Purpose: Allocate memory for 3 dimensional array.
//
// Returns: Pointer to pointer to pointer to doubles
//
// Notes:  Underlying memory is array of matrices (each contiguous).
//    Makes mxFree()'ing easy.
// ********************************************************************/

double ***
mxCalloc3d(int depth, int rows, int cols)
{
   double *tmp;
   double ***tmp3;
   int i,j;
   
   tmp = (double *)mxCalloc(depth*rows*cols,sizeof(double));
   tmp3 = (double***)mxCalloc(depth, sizeof(double **));
   for (i = 0; i < depth; i++)
   {
      tmp3[i] = (double **)mxCalloc(rows,sizeof(double *));
      for (j = 0; j < rows; j++)
         tmp3[i][j] = &tmp[i*rows*cols + j*cols];  /* pointer assignment */
   }
   return(tmp3);
}

/* ********************************************************************
// function: mxCalloc3di
// 
// Purpose: Allocate memory for 3 dimensional array of ints;
//
// Returns: Pointer to pointer to pointer to ints
//
// Notes:  Underlying memory is array of matrices (each contiguous).
//    Makes mxFree()'ing easy.
// ********************************************************************/

int ***
mxCalloc3di(int depth, int rows, int cols)
{
   int *tmp;
   int ***tmp3;
   int i,j;

   tmp = (int *)mxCalloc(depth*rows*cols,sizeof(int));
   tmp3 = (int ***)mxCalloc(depth,sizeof(int **));
   for (i = 0; i < depth; i++)
   {
      tmp3[i] = (int **)mxCalloc(rows,sizeof(int *));
      for (j = 0; j < rows; j++)
         tmp3[i][j] = &tmp[i*rows*cols + j*cols];  /* pointer assignment */
   }
   return(tmp3);
}



/* ********************************************************************
// function: mxCalloc2d
// 
// Purpose: Allocate memory for 2 dimensional matrix.
//
// Returns: Pointer to array of pointers....
//
// Notes:  Underlying matrix is one dimensional, with pointers pointing
// to correct locations.  Makes mxFree()'ing easy.
// ********************************************************************/

double **
mxCalloc2d(int rows, int cols)
{
   double *tmp;
   double **tmp2;
   int i;

   tmp  = (double *)mxCalloc(rows * cols,sizeof(double));
   tmp2 = (double**)mxCalloc(rows,sizeof(double *));
   
   for (i = 0; i < rows; i++)
      tmp2[i] = &tmp[i*cols];  /* pointer assignment */
   return(tmp2);
}

// ********************************************************************
// function: mxCalloc2di
// 
// Purpose: Allocate memory for 2 dimensional matrix.
//
// Returns: Pointer to array of pointers of ints ....
//
// Notes:  Underlying matrix is one dimensional, with pointers pointing
// to correct locations.  Makes mxFree()'ing easy.
// ********************************************************************

int **
mxCalloc2di(int rows, int cols)
{
   int *tmp;
   int **tmp2;
   int i;

   tmp  = (int *)mxCalloc(rows * cols,sizeof(int));
   tmp2 = (int **)mxCalloc(rows,sizeof(int *));
   
   for (i = 0; i < rows; i++)
      tmp2[i] = &tmp[i*cols];  /* pointer assignment */
   return(tmp2);
}

/* ********************************************************************
// function: mxFree3d
// 
// Purpose: Deallocate memory for 3 dimensional array , either double or 
//  float 
//
// Returns: void
//
// Notes:  Use *ONLY* with matrices created with mxCalloc3d() or mxCallocf3d!
// ********************************************************************/
void
mxFree3d(double ***matrix, int depth)
{
   int i;
   
   mxFree(&matrix[0][0][0]);
   for (i = 0; i < depth; i++)
      mxFree(matrix[i]);
   mxFree(matrix);
}

/* ********************************************************************
// function: mxFree3di
// 
// Purpose: Deallocate memory for 3 dimensional array of integers 
//
// Returns: void
// ********************************************************************/
void
mxFree3di(int ***matrix, int depth)
{
   int i;

   mxFree(&matrix[0][0][0]);
   for (i = 0; i < depth; i++)
      mxFree(matrix[i]);
   mxFree(matrix);
}

/* ********************************************************************
// function: mxFree2d
// 
// Purpose: Deallocate memory for 2 dimensional matrix (double or float)
//
// Returns: void
//
// Notes:  Use *ONLY* with matrices created with mxCalloc2d() or mxCallocf2d()!
// *******************************************************************/

void
mxFree2d(double **matrix)
{
   mxFree(matrix[0]);
   mxFree(matrix);
}

/* ********************************************************************
// function: mxFree2di
// 
// Purpose: Deallocate memory for 2 dimensional matrix of integers
//
// Returns: void
//
// Notes:  Use *ONLY* with matrices created with mxCalloc2di() 
// *******************************************************************/

void
mxFree2di(int **matrix)
{
   mxFree(matrix[0]);
   mxFree(matrix);
}

/*
 * Some handy functions for min and max
 */ 
 
double min(double x, double y)
{
  return(x < y ? x : y);
}
double max(double x, double y)
{
  return(x > y ? x : y);
}
double min3(double x, double y, double z)
{
  return(min(x,min(y,z)));
}
double max3(double x, double y, double z)
{
  return(max(x,max(y,z)));
}
/* Return index along with min value */
void min3i(double x, double y, double z, double *minval, int *minind)
{
  if (x < y)
  {
    if (x < z)
    {
      *minval = x;
      *minind = 1;
    }
    else
    {
      *minval = z;
      *minind = 3;
    }
  }
  else if (y < z)
  {
    *minval = y;
    *minind = 2;
  }
  else
  {
    *minval = z;
    *minind = 3;
  }
}
/* Return index along with max value */
void max3i(double x, double y, double z, double *maxval, int *maxind)
{
  if (x > y)
  {
    if (x > z)
    {
      *maxval = x;
      *maxind = 1;
    }
    else
    {
      *maxval = z;
      *maxind = 3;
    }
  }
  else if (y > z)
  {
    *maxval = y;
    *maxind = 2;
  }
  else
  {
    *maxval = z;
    *maxind = 3;
  }
}

/*
 * Solve order preserving assignment problem
 */
void lopap(double **D, int N, int m, int n, double dum_cost, double *Popt, double *T, int *flag)
{

    int i,j,k;      
    int jmin, jmax, kmin, kmax;
    double con1, con2, con3;
    int tmp;
    int ibest, jbest;
    int opt_sz;
    
    extern double ***L;
    extern int ***I;
    extern int ***J; 
    extern int **x;
    extern int minM; 
    extern int maxM;
    extern double *u;
    extern int *v;
    extern double *Tmp;
    extern double *min_cost;
    extern int *Ibest; 
    extern int *Jbest;
    extern int *filledrows;
    extern int *filledcols;
 
    /* Initialize min_cost */
    for (k=0; k<maxM; k++)
      min_cost[k]=HUGE;

    for (i=0; i<m; i++)
      for (j=0; j<n; j++)
      {
        L[i][j][0]=D[i][j];
        if (D[i][j]<min_cost[0])
          min_cost[0]=D[i][j];
      }

    for (k=1; k<maxM; k++)
      min_cost[k]=HUGE;

    /* LOPAP: main loop */
    
    for (i=1; i<m; i++)
    {
     
      jmin=max(1,minM-m+i);
      jmax=min(n-minM+i,n-1);
      for (j=jmin; j<=jmax; j++)
      {
        kmin=max3(1,minM-(m-i),minM-(n-j));
        kmax=min3(i,j,maxM-1);
        for (k=kmin; k<=kmax; k++)
        {
          con1=L[i-1][j-1][k-1];
          if (L[i][j-1][k] < HUGE)
            con2=L[I[i][j-1][k]][J[i][j-1][k]][k-1];
          else
            con2=HUGE;
          if (L[i-1][j][k] < HUGE)
            con3=L[I[i-1][j][k]][J[i-1][j][k]][k-1];
          else
            con3=HUGE;
        

          min3i(con1, con2, con3, u, v);
   
          
          switch (*v)
          {
            case 1:
                I[i][j][k]=i-1;
                J[i][j][k]=j-1;
                break;
            case 2:
                I[i][j][k]=I[i][j-1][k];
                J[i][j][k]=J[i][j-1][k];
                break;
            case 3:
                I[i][j][k]=I[i-1][j][k];
                J[i][j][k]=J[i-1][j][k];
                break;
            default: 
                printf("Switch statement error: *v = %d\n",*v);
                return;
          }
          
            
          L[i][j][k]=*u+D[i][j];
          if (k >= minM-1 & L[i][j][k] < min_cost[k])
          {
            min_cost[k]=L[i][j][k];
            Ibest[k]=i;
            Jbest[k]=j;
          }
        }
      } 
    }
 
 
    
    *Ttmp=HUGE;
    opt_sz=0;  
  
    for (k=minM-1; k<maxM; k++)
    {
      if (min_cost[k]+dum_cost*(N-k-1) < *Ttmp)
      {
        *Ttmp = min_cost[k]+dum_cost*(N-k-1);
        ibest=Ibest[k];
        jbest=Jbest[k];
        opt_sz = k+1;
   
        }
    }
    if (opt_sz == 0)
    {
      printf("Min cost too large.  Exiting.\n");
      return;
    }
  
    
    if (*Ttmp < *T)
    {
      *flag = 1;
      *T = *Ttmp;
     
      /* compute permutation */
        
      /* Represent matching with N by N matrix x: x[i][j] == 1 <=> 
       * point i in first point set matches with point j in second point set */
      for (i=0; i<N; i++)
      {
        for (j=0; j<N; j++) 
        {
          x[i][j]=0;
        }
      }
  
      /* Initialize */
      for (i=0; i<m; i++)
        filledrows[i]=0;
      for (j=0; j<n; j++)
        filledcols[j]=0;
    
      i=ibest;
      j=jbest;
      x[i][j]=1;
      filledrows[i]=1;
      filledcols[j]=1;
      for (k=opt_sz-1; k > 0;  k--)
      {
        tmp=I[i][j][k];
        j=J[i][j][k];
        i=tmp;
        x[i][j]=1;
        filledrows[i]=1;
        filledcols[j]=1;
      }

     
      
      /* Dummies to dummies */
      for (k=0; k<opt_sz-minM; k++)
        x[N-k-1][N-k-1]=1;

      /* Non-dummies to dummies */
      k=0;
      for (i=0; i<m; i++)
      {
        if (filledrows[i] == 0)
        {
          x[i][n+k]=1;
          k++;
        }
        if (k == N-n-(opt_sz-minM))
          break;
      }
  
      /* Dummies to non-dummies */
      k=0;
      for (j=0; j<n; j++)
      {
        if (filledcols[j] == 0)
        {
          x[m+k][j]=1;
          k++;
        }
        if (k == N-m-(opt_sz-minM))
          break;
      }
  
      /* Convert x to permutation vector P */
      for (j=0; j<N; j++)
      {
        i=0;
        while (x[i][j]==0)
          i++;
        Popt[j]=i+1;
      }
    } /* if (*Ttmp < *T) */
    else
      *flag = 0;
    
}

/* ******************************************************************
 * Main routine
 * ****************************************************************** */
void copap(double **C, int N, int m, int n, double dum_cost, int delta, double *P, double *T)
{
  int i,j,k,s;
  int shift, opt_shift, init_opt_shift;
  int found;
  int tmp;
  int *flag;
  double **D;
  double *Popt; 
  
  extern double ***L;
  extern int ***I;
  extern int ***J; 
  extern int **x;
  extern int minM;
  extern int maxM;
  extern double *u;
  extern int *v;
  extern double *Tmp;
  extern double *min_cost;
  extern int *Ibest;
  extern int *Jbest;
  extern int *filledrows;
  extern int *filledcols;

  D = mxCalloc2d(m,n);
  Popt = (double *) mxCalloc(N,sizeof(double));
  flag = (int *) mxCalloc(1,sizeof(int));
  
  *T = HUGE;
  *flag=0;
    
  minM = m+n-N;                
  if (dum_cost > 0)          
    maxM = min(m,n);    
  else
    maxM = minM;

  L = mxCalloc3d(m,n,maxM);
  I = mxCalloc3di(m,n,maxM);
  J = mxCalloc3di(m,n,maxM);
  min_cost= (double *) mxCalloc(maxM,sizeof(double));
  Ibest = (int *) mxCalloc(maxM,sizeof(int));
  Jbest = (int *) mxCalloc(maxM,sizeof(int));
  x = mxCalloc2di(N,N);
  u = (double *) mxCalloc(1,sizeof(double));
  v = (int *) mxCalloc(1,sizeof(int));
  Ttmp = (double *) mxCalloc(1,sizeof(double));
  filledrows = (int *) mxCalloc(m,sizeof(int));
  filledcols = (int *) mxCalloc(n,sizeof(int));
       
  for (i=0; i<m; i++)
    for (j=0; j<n; j++)
      for (k=0; k<maxM; k++)
        L[i][j][k]=HUGE;
  
  /* ***************************************************
   * First pass
   * *************************************************** */
        
  for (shift=0; shift<m; shift+=delta)
  {
    /* Circularly shift cost matrix down by shift rows */
    for (i=0; i<shift; i++)
    {
      for (j=0; j<n; j++)
      {
        D[i][j]=C[m-shift+i][j];
      }
    }
    for (i=shift; i<m; i++)
    {
      for (j=0; j<n; j++)
      {
        D[i][j]=C[i-shift][j];
      }
    }

    
    lopap(D,N,m,n,dum_cost,Popt,T,flag);
  
  
    if (*flag == 1)
      opt_shift = shift;
    
  } /* for shift */
 
  /* ***************************************************
   * Refine search near opt_shift 
   * *************************************************** */
  
  init_opt_shift = opt_shift;
  if (delta > 1)
  {
    for (s=-2*delta+1; s<2*delta; s++)
    {
      if (s==0 || s == -delta || s == delta)
        s++;
      shift = init_opt_shift + s;    
      if (shift < 0)
        shift = m+shift;
      else if (shift >= m)
        shift = shift-m;
    
      /* Circularly shift cost matrix down by shift rows */
      for (i=0; i<shift; i++)
      {
        for (j=0; j<n; j++)
        {
          D[i][j]=C[m-shift+i][j];
        }
      }
      for (i=shift; i<m; i++)
      {
       for (j=0; j<n; j++)
        {
          D[i][j]=C[i-shift][j];
        }
      }
      
      /* Solve linear LOPAP for shifted cost matrix */

      lopap(D,N,m,n,dum_cost,Popt,T,flag);
     
      if (*flag == 1)
        opt_shift = shift;
      
    } /* for s */
  } /* if (delta > 1) */
 
  
  /* Determine P from Popt by compensating for shift */

  for (j=0; j<N; j++)
  {
    found = 0;
    for (i=0; i<m; i++) 
    {
      if (abs(Popt[j] - (i+1)) < .5)          
      {
        found = 1;
        if (i < opt_shift)
          P[j]=m-opt_shift+i+1;    
        else
          P[j]=i-opt_shift+1;    
        break;
      }
    }
    if (!found)
      P[j]=Popt[j];
  }

  mxFree3d(L,m);
  mxFree3di(I,m);
  mxFree3di(J,m);
  mxFree(min_cost);
  mxFree(u);
  mxFree(v);
  mxFree(Ttmp);
  mxFree(Ibest);
  mxFree(Jbest);
  mxFree2di(x);
  mxFree(filledrows);
  mxFree(filledcols);

  mxFree2d(D);
  mxFree(Popt);

  
}

/* ****************************************
 * Mex gateway function for copap
 * **************************************** */
void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[] )
{
  double *origC;    /* Input: square matrix of scores (in vector form) */
  double **C;        /* Scores converted to matrix form */
  int ell;          /* Input: minimum matching size */
  int m, n;         /* number of source/target points; dims of C */
  double dum_cost;  /* Input: cost of assigning to dummy point */
  double *P;           /* Output: optimal assignment */
  double *T;        /* Output: cost of optimal assignment */
  int N;         /* Size of permutation, = m + n - ell */
  int mrows, ncols; /* For error checking */
  int i,j;
  int delta;
  
  /* Check number of arguments */

  if (nrhs != 4 & nrhs != 3)  
    mexErrMsgTxt("Four or three input arguments required."); 
  if (nlhs != 2) 
    mexErrMsgTxt("Two output arguments required."); 
   
  /* Error checking for first argument */
 
  m = mxGetM(prhs[0]);      /* Number of rows */
  n = mxGetN(prhs[0]);      /* Number of cols */
  
  if (!mxIsNumeric(prhs[0])) 
    mexErrMsgTxt("Argument must be numeric.");
  if (mxIsComplex(prhs[0]))
    mexErrMsgTxt("Argument must be real.");
  if (mxGetNumberOfDimensions(prhs[0]) > 2) 
    mexErrMsgTxt("Input must be a two-dimensional matrix");  
  
  /* Error checking for second argument */
    
  mrows = mxGetM(prhs[1]);
  ncols = mxGetN(prhs[1]);
  if(!mxIsDouble(prhs[1]) || mxIsComplex(prhs[1]) || !(mrows == 1 && ncols == 1) ) 
    mexErrMsgTxt("Second input must be a noncomplex scalar integer.");
  
  /* Error checking for third argument */
    
  mrows = mxGetM(prhs[2]);
  ncols = mxGetN(prhs[2]);
  if(!mxIsDouble(prhs[2]) || mxIsComplex(prhs[2]) || !(mrows == 1 && ncols == 1) ) 
    mexErrMsgTxt("Third input must be a positive integer.");
 
  if (nrhs == 4)
    delta = (double) mxGetScalar(prhs[3]);
  else
    delta = 1;
    
    
  /* Assign (pointers to) inputs */
  
  origC = (double *) mxGetPr(prhs[0]);
  ell = (int) mxGetScalar(prhs[1]);
  dum_cost = (double) mxGetScalar(prhs[2]);
  
  /* Additional error checking */
  
  if (ell > m || ell > n) 
    mexErrMsgTxt("Matching size cannot exceed dims of C");
  if (m <=0 || n <= 0)
    mexErrMsgTxt("C is degenerate.");
  if (dum_cost < 0)
    mexErrMsgTxt("Third argument should be non-negative ");
  
  /* Assign pointers to outputs */
  
  N = m + n - ell;
  plhs[0] = mxCreateDoubleMatrix(1,N,mxREAL);
  plhs[1] = mxCreateDoubleMatrix(1,1,mxREAL);
  P = (double *) mxCalloc(N, sizeof(double));
  P = mxGetPr(plhs[0]);
  T = mxGetPr(plhs[1]);

  /* Convert costs from linear array to matrix form */
  
  C = mxCalloc2d(m,n);
  for (i=0; i<m; i++)
    for (j=0; j<n; j++) 
      C[i][j]=origC[i+m*j];
 
  /* Call main routine */
    
  copap(C, N, m, n, dum_cost, delta, P, T);
  
 /* mxFree2d(C);
  mxFree(P);
  */
} /* end of mexFunction */
